import os
import glob
import optparse
import re
import sys
import time

# emdash imports
import emdash.config
import emdash.proxy
import emdash.handlers


def main():
	parser = emdash.config.DBRPCOptions()
	usage = """%prog upload [options] <record type> <record name> <files to upload>

	Record type can be any valid database protocol.
	Some record types have special, application-specific handlers, e.g.:

	ccd			CCD Frames
	scan		Scanned micrographs
	stack		Tomograms

	Other values, e.g. "volume", will create child records of that type, with 1 file per child record.

	Alternatively, you can use "none" for record type and the files will be attached directly to the specified record ID.
	"""

	parser.set_usage(usage)
	parser.add_option("--interactive", action="store_true", dest="interactive", default=True, help="Prompt for parameter values")
	parser.add_option("--noninteractive", action="store_false", dest="interactive", help="Do not prompt for parameter values")
	parser.add_option("--force", action="store_true", help="Force re-upload even if a sidecar is found")
	parser.add_option("--metafile", action="store_true", dest="metafile", help="Attempt to read JAMES/JADAS metadata files (default)", default=True)
	parser.add_option("--nometafile", action="store_false", dest="metafile", help="Ignore metadata files")
	(options, args) = parser.parse_args()

	if len(args) < 3:
		parser.error("Record type, Record name, and files to upload are required")

	rectype = args[0]
	name = args[1]
	filenames = args[2:]

	try:
		int(name)
	except ValueError:
		print "Invalid Record name '%s' -- see usage."%name
		return

	print "%s Files to upload:"%len(filenames)
	for i in filenames:
		print "\t",i


	db = parser.opendb(request=True)


	handlerclass = emdash.handlers.gethandler(rectype=rectype)
	
	applyparams = {}
	if options.interactive:
		applyparams = get_param_values(handlerclass.request_params, db=db)

	dbts = []
	filecount = len(filenames)
	for count, filename in enumerate(filenames):
		dbt = handlerclass(
			db=db,
			rectype=rectype,
			filename=filename,
			name=name,
			applyparams=applyparams,
			pos=(count,filecount)
		)
		dbt.upload()




#########################
# Ugly console input code
#########################

def get_param_values(params, db=None):
	pds = db.getparamdef(params)
	pds = dict([[i.get('name'), i] for i in pds])
	
	ret = {}
	for param in params:
		pd = pds.get(param)
		ret[param] = get_param_value_console(pd, db=db)
	return ret



def get_param_value_console(param, db=None):
	name = param.get('name')

	print "\n----- %s -----"%name
	print "\tDescription: %s"%param.get('desc_long')
	if param.get('defaultunits'):
		print "\tUnits: %s"%param.get('defaultunits')


	punits = param.get('defaultunits') or ''
	lines = None
	selections = []

	try:
		lines = [i[0] for i in db.findvalue(name, '', True, False, 5)]
	except Exception, inst:
		# print inst
		pass

	count = 0
	if param.get('choices'):
		print "\n\tSuggested values:"
		for choice in param.get('choices', []):
			print "\t\t%s) %s"%(count, choice)
			selections.append(choice)
			count += 1

	if param.get('vartype') != "choice" and lines:
		if param.get('choices'):
			print "\n\tOther common values:"
		else:
			print "\n\tCommon values:"

		for line in lines:
			print "\t\t%s) %s"%(count, line)
			selections.append(line)
			count += 1

	print "\n\t\t%s) None or N/A"%count
	selections.append('')
	count += 1

	if param.get('vartype') != "choice":
		print "\t\t%s) Enter a different not listed above"%count


	while True:
		inp = raw_input("\n\tSelection (0-%s): "%count)
		try:
			inp = int(inp)

			if inp == count:
				inp = raw_input("\tValue: ")
			else:
				inp = selections[inp]
		except:
			continue

		break

	return inp


if __name__ == "__main__":
	main()
