# $Id: p41.py,v 1.36 2011/04/25 16:45:01 irees Exp $
import cStringIO
import collections

import emen2.util.listops
import emen2.db.config
g = emen2.db.config.g()
from emen2.web.view import View

def sort_grant_source(pub):
	if pub["grant_source"] == "NIH": return 2
	if pub["grant_source_type"] == "USA_Federal": return 1
	return 0



@View.register
class P41Report(View):

	@View.add_matcher(r'^/record/(?P<name>\d+)/p41report/$', r'^/record/(?P<name>\d+)/p41report/(?P<action>.+)/$')
	def init(self, name=None, action=None, **kwargs):

		self.template = '/pages/p41'
		if action == 'years':
			self.template = '/pages/p41.years'
		elif action == 'xml':
			self.template = '/pages/p41.xml'
			self.mimetype = 'text/xml'


		# Get the p41_grant_info
		root = self.db.getrecord(name)
		year = root.get('p41_year')
		print "Building report for %s"%year


		# Get the p41 projects
		names = set()
		p41_names = self.db.getchildren(name, rectype="p41_project")
		p41_projects = self.db.getrecord(p41_names)
		print "Total p41 projects for %s: %s"%(year, len(p41_projects))


		# Referenced users
		users = set(root['grant_pi'])
		usersbyproject = {}
		for rec in p41_projects:
			u = set([rec['p41_project_pi']] + rec.get('p41_project_investigators', []))
			if len(u) == 0:
				print "WARNING! Project %s has no investigators"%rec.name
			usersbyproject[rec.name] = u
			users |= u


		# Store this as a dict to make referencing easier
		users = self.db.getuser(users)
		for user in users:
			user.getdisplayname(lnf=True)
		users = emen2.util.listops.dictbykey(users, 'name')
		print "Total users for %s: %s"%(year, len(users))


		# Grants are children of a user, and parents of a p41 project
		grantsbyuser = collections.defaultdict(list)
		projectsbygrant = collections.defaultdict(set)
		grants = set()
		for user in users.values():
			print "Grants for user %s:"%(user.name)
			g = self.db.getrecord(self.db.getchildren(user.record, rectype="grant"))
			g_expired = filter(lambda x:int(x['grant_p41_ref_end'][:4]) < year, g)
			g = filter(lambda x:int(x['grant_p41_ref_end'][:4]) >= year, g)
			print "\tCurrent: %s, Expired: %s"%(len(g), len(g_expired))

			for grant in g:
				print "\tGrant %s"%grant.name
				v = self.db.getchildren(grant.name, rectype='p41_project') & p41_names
				if len(v) == 0:
					print "\t\tDoes not reference any current projects..."

				for v2 in v:
					if user.name not in usersbyproject[v2]:
						print "\t\tUser is not referneced in project %s"%v2
					else:
						print "\t\tReferences: %s"%v2
						projectsbygrant[grant.name].add(v2)
						grants.add(grant.name)
						if grant.name not in [j.name for j in grantsbyuser[user.name]]:
							grantsbyuser[user.name].append(grant)
			
			grantsbyuser[user.name] = sorted(grantsbyuser[user.name], key=sort_grant_source)


		print "Total included grants for %s: %s"%(year, len(grants))


		# Back up each p41 project to find the associated publications.
		pubsbyproject = {}
		pubs = set()
		parents = self.db.getparents([i.name for i in p41_projects], rectype="project*")
		for k, v in parents.items():
			print "Finding publications for %s"%k
			p = self.db.getchildren(v, rectype="publication*", recurse=-1)
			if p.values():
				p = emen2.util.listops.combine(*p.values())
			p = self.db.getrecord(p) or []
			p_expired = filter(lambda x:x.get('p41_year')!=year, p)
			p = filter(lambda x:x.get('p41_year')==year and x.get('p41_acknowledged'), p)
			print "\tCurrent: %s, Other years: %s"%(len(p), len(p_expired))
			p = set([i.name for i in p])
			pubsbyproject[k] = p
			pubs |= p
		
		recnames = self.db.renderview(p41_names | grants | pubs)
		rendered = self.db.renderview(p41_names | grants, viewtype="defaultview", edit=True)
		# rendered.update(self.db.renderview(pubs))

		pubs = self.db.getrecord(pubs)
		grants = self.db.getrecord(grants)
		print "Total publications for %s: %s"%(year, len(pubs))

		# Used by all templates
		self.set_context_item('root',root)
		self.set_context_item('year',year)
		self.set_context_item('users',users)
		self.set_context_item('p41_projects',p41_projects)
		self.set_context_item('pubs',pubs)
		self.set_context_item('grants',grants)
		self.set_context_item('grantsbyuser',grantsbyuser)
		self.set_context_item('projectsbygrant',projectsbygrant)
		self.set_context_item('pubsbyproject',pubsbyproject)

		# These are used by the editing interface
		self.set_context_item('usersbyproject',usersbyproject)
		self.set_context_item('recnames',recnames)
		self.set_context_item('rendered',rendered)




__version__ = "$Revision: 1.36 $".split(":")[1][:-1].strip()
